function thresholds = joint_calibration(scores, labels, version, time)

%    Copyright (C) 2015  Davide Modolo
%
%    You can redistribute and/or modify this software for non-commercial use
%    under the terms of the GNU General Public License as published by
%    the Free Software Foundation, either version 3 of the License, or
%    (at your option) any later version.
%
%    This program is distributed in the hope that it will be useful,
%    but WITHOUT ANY WARRANTY; without even the implied warranty of
%    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
%    GNU General Public License for more details.
%
%    You should have received a copy of the GNU General Public License
%    along with this program.  If not, see <http://www.gnu.org/licenses/>.
%
%    For commercial use, contact the author for licensing options.
%
%    Contact: d.modolo@sms.ed.ac.uk

% INPUTS:
%       scores: matrix W x E, where E = number of classifiers to calibrate
%               W = number of windows in the validation set
%                        remind now eesvms scoring on their own
%       labels: window labels (1 if the window is positive, 0 otherwise)
%
%       version: can either be GLOBAL (default) or APPROXIMATE
%
% OUTPUT:
%       thresholds: the thresholds learnt by our joint calibration
%                   procedure (theta in Eq. 2 of the paper)

if nargin < 4
    time = 3600;   % seconds
end

if nargin < 3 
    version = 'GLOBAL';
end

if nargin < 2
    error('Inputs needed: 2, Inputs given: %d', nargin);
end


%% -------------------------- PRE-PROCESSING --------------------------
% Here we pre-compute lots of matrixes/data to reduce the search tree to a
% simple look-up job. These are saved in "data"
fprintf('PRE-PROCESSING STARTED\n');
data.nExemplar = size(scores, 2);
data.nPosW = uint32(sum(labels));
data.nNegW = uint32(sum(~labels));
[~, i_sn] = sort(scores(~labels, :), 'descend');
data.i_sn = uint32(i_sn);

% Select candidate thresholds. 
% 'candidate_thresholds' = matrix (P x E). since #T < P, many cells of this 
% matrix are set to -Inf. 'candidate_thresholds' is sorted descend, column wise.
% This operation corresponds to Sec. 4.1, Fig. 3 of the paper
fprintf('> Computing set of discrete thresholds\n');
data.candidate_thresholds = estimateCandidateThresholds(scores, labels);


% Compute thresholds for positive windows matrix
% Each cell (p, e) of this matrix tells us what is the highest threshold 
% for exemplar 'e' so that positive windows 'p' is scored positive
fprintf('> Retrieving thighest thresholds for each eesvm for each postive window\n');
data.t4p = getTightestThreshold_PerPositive(data.candidate_thresholds,  scores(labels, :));
data.t4p = single(data.t4p);
    

% Count #false positives produced by each threshold for each eesvm
fprintf('> Computing #false positives produced by each threshold of each eesvm\n');
[s_n, ~] = sort(scores(~labels, :), 'descend');
s_n = single(s_n);
data.fp_thresh = computeNFalsePositives_PerThreshold(data.candidate_thresholds, s_n);

% compute upper bound -------------------> get current best solution
fprintf('> Computing initial upper bound:  ');

[data.upper_bound, data.false_pos_count, data.set_negs, data.thresholds_of_UB] = ...
    computeInitialUpperBound(data, scores(~labels, :));
fprintf('%d\n', data.upper_bound);


% Some thresholds will never be in the global optimal solution (since their
% retrieved # of false positives already exceed the upper bound)
% This operation (partially) corresponds to Sec. 4.3, Observation 1. of the paper
thresholds_to_discard = data.fp_thresh > data.upper_bound;

% Compute matrix 'tree_edges'
% In this matrix, each cell tells us if exemplar 'e' can retrieve 'p' 
% 1 = it can, 0 = it cant' (= subtree doesn't need to be searched)
data.tree_edges = zeros(size(data.t4p));
for i = 1:data.nExemplar
    A = find(thresholds_to_discard(:, i) == 1);
    data.tree_edges(:, i) = data.t4p(:, i) < A(1);
end 

% Check what positive windows are retrieved by setting the thresholds for
% all eesvm to the max ones and update tree_edges accordingly.
% This operation corresponds to Sec. 4.3, Observation 3. of the paper
[min_t4p, i_min_t4p] = min(data.t4p, [], 2);
pos_retrieved_a_priori = min_t4p == 1;

data.tree_edges(pos_retrieved_a_priori, :) = 0;
for i = 1:size(data.tree_edges, 1)
    if pos_retrieved_a_priori(i)
        data.tree_edges(i, i_min_t4p(i)) = 1;
    end
end


% Look for equivalences according to Sec. 4.3, Observation 2. of the paper. 
% Note how some equivalences can be discovered before starting the search
fprintf('> Checking for equivalences:   ');
equiv = check4Equivalences(data);

% Eliminate equivalent edges in the tree
for i = 1:data.nPosW
    if ~isempty(equiv{i})
        for j = 1:size(equiv{i}, 1)
            if data.tree_edges(i, equiv{i}(j, 1)) == 1
                data.tree_edges(i, equiv{i}(j, 2)) = 0;
            end   
        end
    end 
end


% Sort the positives according to Sec. 4.3, Observation 4. of the paper.
% We sort positives by highest number of false positives
[s_fp, i_s_fp] = sort(min(data.false_pos_count, [], 2), 'descend');
data.layers_pos_windows = i_s_fp;

%% --------------------------- SEARCH TREE ---------------------------

fprintf('SEARCH STARTED\n');
fprintf('> Current minimum energy: ');
[thresholds, energy] = searchTree(data, version, time);
tic
%fprintf('\n\n-------\n\nFound optimal solution in %.2f seconds\n', toc);
fprintf('\n> Final Minimum Energy: %d\n', energy);


end