function [thresholds, min_energy_so_far] = searchTree(data, version, time)

%    Copyright (C) 2015  Davide Modolo
%
%    You can redistribute and/or modify this software for non-commercial use
%    under the terms of the GNU General Public License as published by
%    the Free Software Foundation, either version 3 of the License, or
%    (at your option) any later version.
%
%    This program is distributed in the hope that it will be useful,
%    but WITHOUT ANY WARRANTY; without even the implied warranty of
%    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
%    GNU General Public License for more details.
%
%    You should have received a copy of the GNU General Public License
%    along with this program.  If not, see <http://www.gnu.org/licenses/>.
%
%    For commercial use, contact the author for licensing options.
%
%    Contact: d.modolo@sms.ed.ac.uk

nodes(1).layer = 0;
nodes(1).negs = uint32([]);
nodes(1).thresholds = data.candidate_thresholds(1, :); 
nodes(1).energy = 0;

min_energy_so_far = data.upper_bound + 1;  % +1 needed for (rare) a corner case
thresholds = nodes(1).thresholds; 

th = tic;   % needed for approximate

% Loop until no nodes remain
while ~isempty(nodes)
    
    if strcmp(version, 'APPROXIMATE') && time_expired(th, time)
        break;
    end

    % -- SELECT NEXT NODE
    % depth first 
    depths = horzcat(nodes(:).layer);
    deepest_nodes = find(depths == max(depths));
    
    % if many nodes with same depth, go for the one with the min energy
    if length(deepest_nodes) == 1
        next_node = deepest_nodes(1);
    else
        energies = horzcat(nodes(deepest_nodes).energy);
        min_E = find(energies == min(energies));
        next_node = deepest_nodes(min_E(1));
    end
    
    currentNode = nodes(next_node);
    nodes(next_node) = [];

    % select the positive window to evaluate next
    next_pos_window = data.layers_pos_windows(currentNode.layer + 1);
     
    % check what eesvm are candidate "takers" of this positive
    new_branches = find(data.tree_edges(next_pos_window, :) == 1);
      
    all_new_nodes = [];
    % we add all of them to the queue  
    for i = 1:length(new_branches)
 
        % initialize new node to parent's params and change its depth
        new_node = currentNode;
        new_node.layer = new_node.layer + 1; 
        
        % check if we need to move the threshold of this eesvm in order to
        % score 'next_pos_window' as positive
        if new_node.thresholds(new_branches(i)) > ...
                data.candidate_thresholds(data.t4p(next_pos_window, ...
                new_branches(i)), new_branches(i))
            
            % assign new threhsold
            new_node.thresholds(new_branches(i)) = ...
                data.candidate_thresholds(data.t4p(next_pos_window, ...
                new_branches(i)), new_branches(i));
              
            % compute new energy 
            new_threshold_negs = ... 
                data.i_sn(1:data.fp_thresh(data.t4p(next_pos_window, ...
                new_branches(i)), new_branches(i)), new_branches(i));
            
            %new_node.negs = unique([new_threshold_negs; new_node.negs]);
            % faster version of unique              
            new_node.negs = fast_unique(new_threshold_negs, new_node.negs, data.nNegW);
            new_node.energy = length(new_node.negs);        
        else  
            % if we don't need to move its threshold, its energy won't
            % change -> go for this one and forget all the other branches 
            all_new_nodes = new_node;
            break
        end  
        
        % if we moved the threshold, but the energy didn't increase -> go
        % for this eesvm and forget all the other branches
        % this happens when the negative windows 'collected' by moving the
        % threshold had already been seen 
        if new_node.energy == currentNode.energy 
            all_new_nodes = new_node;
            break
        end
                             
        % keep the node if the new energy is lower than the current upper bound
        if new_node.energy < min_energy_so_far
            all_new_nodes = [all_new_nodes, new_node];                
        end
        
    end 
 
     
    % check if we have reached a leaf node
    % if so, check 
    if (currentNode.layer + 1) == length(data.layers_pos_windows) && ~isempty(all_new_nodes)
        
        % check what node has the minimum energy
        [min_energy, i_min_energy] =  min(horzcat(all_new_nodes(:).energy));
         
        % update the upper bound and the current best thresholds
        if min_energy(1) < min_energy_so_far
            min_energy_so_far = min_energy(1);
            thresholds = all_new_nodes(i_min_energy(1)).thresholds;
            % fprintf('New lower energy: %d\n', min_energy_so_far);
            fprintf('- %d ', min_energy_so_far);
            
            th = tic;
            
            % eliminate edges with energy grater than the new optimal
            thresholds_to_discard = data.fp_thresh > min_energy_so_far;
            for i = 1:data.nExemplar
                A = find(thresholds_to_discard(:, i) == 1);
                XX = data.t4p(:, i) < A(1);
                data.tree_edges(XX==0, i) = 0;
            end 
            
            % eliminate nodes in queue with energy greater than the new
            % optimal
            nodes_to_discard = zeros(length(nodes),1);
            for i = 1:length(nodes)
                nodes_to_discard(i) = length(nodes(i).negs);
            end
            n2d = find(nodes_to_discard > min_energy_so_far);
            nodes(n2d) = [];
            
%             display(thresholds)
%         else 
%             fprintf('--- Not optimal solution, energy: %d\n', min_energy(1));
%             display(all_new_nodes(i_min_energy(1)).thresholds)
        end
         
        continue;         
    end 
    
    % if the nodes still survived, add them to the queue 
    % sorted by decreasing energy   
    if ~isempty(all_new_nodes)
        [min_energy, i_min_energy] =  sort(horzcat(all_new_nodes(:).energy), 'ascend');
        nodes = [nodes all_new_nodes(i_min_energy(1:min(length(all_new_nodes), 20)))];   
    end
    
end

end